import argparse
import logging
import os
import pickle

from pandas.api.types import is_categorical_dtype

from gama import GamaClassifier, GamaRegressor
from gama.data_loading import X_y_from_file


def parse_args():
    desc = "An AutoML tool that optimizes machine learning pipelines for your data."
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument(
        "input_file",
        type=str,
        help="A csv or ARFF file with the data to optimize a model for.",
    )

    io_group = parser.add_argument_group("File I/O")

    io_group.add_argument(
        "-sep",
        dest="separator",
        type=str,
        default=None,
        help=(
            "For CSV files: the character used to separate values in the input file."
            "If none is given, the Python parser will be used to infer it."
        ),
    )
    io_group.add_argument(
        "--target",
        dest="target",
        type=str,
        default=None,
        help="The target column in the input file.",
    )
    io_group.add_argument(
        "-o",
        dest="output_file",
        type=str,
        default="gama_model.pkl",
        help="Path to store the final model in. (default=gama_model.pkl)",
    )

    io_group.add_argument(
        "-py",
        dest="export_python",
        type=str,
        default=None,
        help="If set, export python code for the final model to this destination.",
    )

    optimization = parser.add_argument_group("Optimization")
    optimization.add_argument(
        "--mode",
        dest="mode",
        type=str,
        default=None,
        help="By default, it is determined automatically from the input file."
        "Use this parameter force 'classification' or 'regression' mode. ",
    )
    optimization.add_argument(
        "-m",
        dest="metric",
        type=str,
        default=None,
        help="The metric to optimize the model for. "
        "Default is log loss for classification and RMSE for regression.",
    )
    optimization.add_argument(
        "--long",
        dest="prefer_short",
        action="store_false",
        help="By default GAMA will guide search towards shorter pipelines. "
        "Set this flag to disable the feature.",
    )

    resources = parser.add_argument_group("Resources")
    resources.add_argument(
        "-t",
        dest="time_limit_m",
        type=int,
        default=60,
        help="The maximum time (in minutes) GAMA has for optimization. (default=60)",
    )
    resources.add_argument(
        "--time_pipeline",
        dest="max_eval_time_m",
        type=int,
        default=5,
        help="The maximum time in minutes GAMA may spend "
        "on a single pipeline evaluation. (default=5)",
    )
    resources.add_argument(
        "-n",
        dest="n_jobs",
        type=int,
        default=1,
        help="Jobs to run in parallel for pipeline evaluations. (default=1)",
    )

    # Extra
    parser.add_argument(
        "-outdir",
        dest="outdir",
        default=None,
        type=str,
        help="Directory to store GAMA logs",
    )
    parser.add_argument(
        "-v",
        dest="verbose",
        action="store_true",
        help="Report status updates to console.",
    )
    parser.add_argument(
        "-dry",
        dest="dry_run",
        action="store_true",
        help="If True, execute without calling fit or exports.",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    print("CLI: Processing input")
    if not os.path.exists(args.input_file.lower()):
        raise FileNotFoundError(args.input_file)
    if args.input_file.lower().split(".")[-1] not in ["csv", "arff"]:
        raise ValueError("Unknown file extension. Please use csv or arff.")

    kwargs = {}
    if args.input_file.lower().endswith(".csv") and args.separator is not None:
        kwargs["sep"] = args.seperator

    x, y = X_y_from_file(
        file_path=args.input_file.lower(), split_column=args.target, **kwargs,
    )
    if args.mode is None:
        if is_categorical_dtype(y.dtype):
            args.mode = "classification"
        else:
            args.mode = "regression"
        print(f"Detected a {args.mode} problem.")

    print("CLI: Initializing GAMA")
    log_level = logging.INFO if args.verbose else logging.WARNING
    configuration = dict(
        regularize_length=args.prefer_short,
        max_total_time=args.time_limit_m * 60,
        max_eval_time=args.max_eval_time_m * 60,
        n_jobs=args.n_jobs,
        verbosity=log_level,
        output_directory=args.outdir,
        store="nothing" if args.dry_run else "logs",
    )
    if args.metric:
        configuration["scoring"] = args.metric

    if args.mode == "regression":
        automl = GamaRegressor(**configuration)
    elif args.mode == "classification":
        automl = GamaClassifier(**configuration)
    else:
        raise ValueError(f"Mode {args.mode} is not valid (--mode).")

    if not args.dry_run:
        print("CLI: Starting model search")
        automl.fit(x, y)

        # == Model Export ===
        print("CLI: Exporting models.")
        with open(args.output_file, "wb") as fh:
            pickle.dump(automl.model, fh)

        if args.export_python is not None:
            automl.export_script(args.export_python, raise_if_exists=False)
    else:
        automl.cleanup("all")
    print("done!")


if __name__ == "__main__":
    main()



from collections import defaultdict
from typing import Dict, Any

import sklearn

from gama.genetic_programming.components import Primitive, Terminal, DATA_TERMINAL


def pset_from_config(configuration):
    """ Create a pset for the given configuration dictionary.

    Given a configuration dictionary specifying operators (e.g. sklearn
    estimators), their hyperparameters and values for each hyperparameter,
    create a gp.PrimitiveSetTyped that contains:

        - For each operator a primitive
        - For each possible hyperparameter-value combination a unique terminal

    Side effect: Imports the classes of each primitive.

    Returns the given Pset.
    """

    pset = defaultdict(list)
    parameter_checks = {}

    # Make sure the str-keys are evaluated first, they describe shared hyperparameters.
    # Order-preserving dictionaries are not in the Python 3.6 specification.
    sorted_keys = reversed(sorted(configuration.keys(), key=lambda x: str(type(x))))
    for key in sorted_keys:
        values = configuration[key]
        if isinstance(key, str):
            # Specification of shared hyperparameters
            for value in values:
                pset[key].append(Terminal(value=value, output=key, identifier=key))
        elif isinstance(key, object):
            # Specification of operator (learner, preprocessor)
            hyperparameter_types = []
            for name, param_values in sorted(values.items()):
                # We construct a new type for each hyperparameter, so we can specify
                # it as terminal type, making sure it matches with expected
                # input of the operators. Moreover it automatically makes sure that
                # crossover only happens between same hyperparameters.
                if isinstance(param_values, list) and not param_values:
                    # An empty list indicates a shared hyperparameter
                    hyperparameter_types.append(name)
                elif name == "param_check":
                    # This allows users to define illegal hyperparameter combinations,
                    # but is not a terminal.
                    parameter_checks[key.__name__] = param_values[0]
                else:
                    hp_name = f"{key.__name__}.{name}"
                    hyperparameter_types.append(hp_name)
                    for value in param_values:
                        pset[hp_name].append(
                            Terminal(value=value, output=name, identifier=hp_name,)
                        )

            # After registering the hyperparameter types,
            # we can register the operator itself.
            transformer_tags = [
                "DATA_PREPROCESSING",
                "FEATURE_SELECTION",
                "DATA_TRANSFORMATION",
            ]
            if issubclass(key, sklearn.base.TransformerMixin) or (
                hasattr(key, "metadata")
                and key.metadata.query()["primitive_family"] in transformer_tags
            ):
                pset[DATA_TERMINAL].append(
                    Primitive(
                        input=hyperparameter_types, output=DATA_TERMINAL, identifier=key
                    )
                )
            elif issubclass(key, sklearn.base.ClassifierMixin) or (
                hasattr(key, "metadata")
                and key.metadata.query()["primitive_family"] == "CLASSIFICATION"
            ):
                pset["prediction"].append(
                    Primitive(
                        input=hyperparameter_types, output="prediction", identifier=key
                    )
                )
            elif issubclass(key, sklearn.base.RegressorMixin) or (
                hasattr(key, "metadata")
                and key.metadata.query()["primitive_family"] == "REGRESSION"
            ):
                pset["prediction"].append(
                    Primitive(
                        input=hyperparameter_types, output="prediction", identifier=key
                    )
                )
            else:
                raise TypeError(
                    f"Expected {key} to be either subclass of "
                    "TransformerMixin, RegressorMixin or ClassifierMixin."
                )
        else:
            raise TypeError(
                "Encountered unknown type as key in dictionary."
                "Keys in the configuration should be str or class."
            )

    return pset, parameter_checks


def merge_configurations(c1, c2):
    """ Takes two configurations and merges them together. """
    # Should refactor out 6 indentation levels
    merged: Dict[Any, Any] = defaultdict(lambda: None, c1)
    for algorithm, hparams2 in c2.items():
        if algorithm not in merged:
            merged[algorithm] = hparams2
            continue

        hparams = merged[algorithm]
        if isinstance(hparams, list) and isinstance(hparams2, list):
            merged[algorithm] = list(set(hparams + hparams2))
            continue  # Here the algorithm is actually a shared hyperparameter.

        for hyperparameter, values in hparams2.items():
            if hyperparameter not in hparams:
                hparams[hyperparameter] = values
                continue  # Hyperparameter only specified in one configuration.

            values1 = hparams[hyperparameter]
            if isinstance(values1, dict) and isinstance(values, dict):
                hparams[hyperparameter] = {**values1, **values}
            elif isinstance(values1, type(values)):
                # Both are ranges, arrays or lists.
                hparams[hyperparameter] = list(set(list(values1) + list(values)))
            else:
                raise TypeError(
                    f"Could not merge values of {algorithm}.{hyperparameter}:"
                    f"{hparams} vs. {hparams2}"
                )
    return merged



from abc import ABC
from typing import List, Dict, Tuple, Any, Union

import pandas as pd

from gama.genetic_programming.operator_set import OperatorSet
from gama.genetic_programming.components import Individual
from gama.logging.evaluation_logger import EvaluationLogger


class BaseSearch(ABC):
    """ All search methods should be derived from this class.
    This class should not be directly used to configure GAMA.
    """

    def __init__(self):
        # hyperparameters can be used to safe/process search hyperparameters
        self._hyperparameters: Dict[str, Tuple[Any, Any]] = dict()
        self.output: List[Individual] = []
        self.logger = EvaluationLogger

    def __str__(self):
        # Not sure if I should report actual used hyperparameters (i.e. include default)
        # or only those set by user.
        user_set_hps = {
            parameter: set_value
            for parameter, (set_value, default) in self._hyperparameters.items()
            if set_value is not None
        }
        hp_configuration = ",".join(
            [f"{name}={value}" for (name, value) in user_set_hps.items()]
        )
        return f"{self.__class__.__name__}({hp_configuration})"

    @property
    def hyperparameters(self) -> Dict[str, Any]:
        """ Hyperparameter (name, value) pairs as set/determined dynamically/default.

         Values may have been set directly, through dynamic defaults or static defaults.
         This is also the order in which the value of a hyperparameter is checked,
         i.e. a user set value wil overwrite any other value, and a dynamic default
         will overwrite a static one.
         Dynamic default values only considered if `dynamic_defaults` has been called.
         """
        return {
            parameter: set_value if set_value is not None else default
            for parameter, (set_value, default) in self._hyperparameters.items()
        }

    def _overwrite_hyperparameter_default(self, hyperparameter: str, value: Any):
        set_value, default_value = self._hyperparameters[hyperparameter]
        self._hyperparameters[hyperparameter] = (set_value, value)

    def dynamic_defaults(
        self, x: pd.DataFrame, y: Union[pd.DataFrame, pd.Series], time_limit: float
    ) -> None:
        """ Set hyperparameter defaults based on the dataset and time-constraints.

        Should be called before `search`.

        Parameters
        ----------
        x: pandas.DataFrame
            Features of the data.
        y: pandas.DataFrame or pandas.Series
            Labels of the data.
        time_limit: float
            Time in seconds available for search and selecting dynamic defaults.
            There is no need to adhere to this explicitly,
            a `stopit.utils.TimeoutException` will be raised.
            The time-limit might be an important factor in setting hyperparameter values
        """
        # updates self.hyperparameters defaults
        raise NotImplementedError("Must be implemented by child class.")

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        """ Execute search as configured.

        Sets `output` field of this class to the best Individuals.

        Parameters
        ----------
        operations: OperatorSet
            Has methods to create new individuals, evaluate individuals and more.
        start_candidates: List[Individual]
            A list of individuals to be considered before all others.
        """
        raise NotImplementedError("Must be implemented by child class.")


def _check_base_search_hyperparameters(
    toolbox, output: List[Individual], start_candidates: List[Individual]
) -> None:
    """ Checks that search hyperparameters are valid.

    :param toolbox:
    :param output:
    :param start_candidates:
    :return:
    """
    if not isinstance(start_candidates, list):
        raise TypeError(
            f"'start_population' must be a list but was {type(start_candidates)}"
        )
    if not all(isinstance(x, Individual) for x in start_candidates):
        raise TypeError(f"Each element in 'start_population' must be Individual.")



from datetime import datetime
from functools import partial
import operator
from typing import Optional, Dict, Callable, Iterable

from gama.logging import TIME_FORMAT
from gama.utilities.evaluation_library import Evaluation


def nested_getattr(o, attr):
    for a in attr.split("."):
        o = getattr(o, a)
    return o


class EvaluationLogger:
    def __init__(
        self,
        file_path: str,
        separator: str = ";",
        fields: Optional[Dict[str, Callable[[Evaluation], str]]] = None,
        extra_fields: Optional[Dict[str, Callable[[Evaluation], str]]] = None,
    ):
        """ Formats evaluations for output to a csv file.

        Parameters
        ----------
        file_path: str
            The log file to write to.
        separator: str (default=';')
            The delimiter for the csv file.
            Note that the default `fields` results in ',' is cell values.
        fields: Dict[str, Callable[[Evaluation], str]], optional (default=None)
            Mapping of column names to a function which extracts the corresponding
            value from an evaluation.
            If None, a default set of columns is used.
        extra_fields: Dict[str, Callable[[Evaluation], str]], optional (default=None)
            Additional fields to log. Useful if you want to keep the default `fields`,
            but need additional information.
        """
        self._file_path = file_path
        self._sep = separator

        if fields is None:
            self.fields: Dict[str, Callable[[Evaluation], str]] = dict(
                id=partial(nested_getattr, attr="individual._id"),
                pid=operator.attrgetter("pid"),
                t_start=partial(nested_getattr, attr="individual.fitness.start_time"),
                t_wallclock=partial(
                    nested_getattr, attr="individual.fitness.wallclock_time"
                ),
                t_process=partial(
                    nested_getattr, attr="individual.fitness.process_time"
                ),
                score=partial(nested_getattr, attr="individual.fitness.values"),
                pipeline=lambda e: e.individual.pipeline_str(),
                error=operator.attrgetter("error"),
            )
        else:
            self.fields = fields

        if extra_fields is not None:
            self.fields.update(extra_fields)

        self.log_line(list(self.fields))

    def log_line(self, values: Iterable[str]):
        """ Appends `values` as a row of separated values to the file. """
        with open(self._file_path, "a") as evaluations:
            evaluations.write(self._sep.join(values) + "\n")

    def log_evaluation(self, evaluation):
        values = [getter(evaluation) for getter in self.fields.values()]

        def format_value(v):
            if isinstance(v, datetime):
                return v.strftime(TIME_FORMAT)
            return str(v)

        self.log_line(map(format_value, values))



import shutil
from abc import ABC
from collections import defaultdict
from functools import partial, partialmethod
import logging
import multiprocessing
import os
import random
import subprocess
import time
import uuid
from typing import (
    Union,
    Tuple,
    Optional,
    Dict,
    Type,
    List,
    Callable,
    Any,
    Iterable,
    cast,
)
import warnings

import pandas as pd
import numpy as np
import stopit
from sklearn.base import TransformerMixin
from sklearn.pipeline import Pipeline

import gama.genetic_programming.compilers.scikitlearn
from gama.genetic_programming.components import Individual, Fitness
from gama.search_methods.base_search import BaseSearch
from gama.utilities.evaluation_library import EvaluationLibrary, Evaluation
from gama.utilities.metrics import scoring_to_metric

from gama.__version__ import __version__
from gama.data_loading import X_y_from_file
from gama.data_formatting import format_x_y
from gama.search_methods.async_ea import AsyncEA
from gama.utilities.generic.timekeeper import TimeKeeper
from gama.logging.utility_functions import register_stream_log
from gama.utilities.preprocessing import (
    basic_encoding,
    basic_pipeline_extension,
)
from gama.genetic_programming.mutation import random_valid_mutation_in_place
from gama.genetic_programming.crossover import random_crossover
from gama.genetic_programming.selection import (
    create_from_population,
    eliminate_from_pareto,
)
from gama.genetic_programming.operations import create_random_expression
from gama.configuration.parser import pset_from_config
from gama.genetic_programming.operator_set import OperatorSet
from gama.genetic_programming.compilers.scikitlearn import compile_individual
from gama.postprocessing import (
    BestFitPostProcessing,
    BasePostProcessing,
    EnsemblePostProcessing,
)
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from gama.utilities.metrics import Metric

log = logging.getLogger(__name__)

STR_NO_OPTIMAL_PIPELINE = """Gama did not yet establish an optimal pipeline.
                          This can be because `fit` was not yet called, or
                          did not terminate successfully."""

for module_to_ignore in ["sklearn", "numpy"]:
    warnings.filterwarnings("ignore", module=module_to_ignore)


class Gama(ABC):
    """ Wrapper for the toolbox logic surrounding executing the AutoML pipeline. """

    def __init__(
        self,
        scoring: Union[
            str, Metric, Iterable[str], Iterable[Metric]
        ] = "filled_in_by_child_class",
        regularize_length: bool = True,
        max_pipeline_length: Optional[int] = None,
        config: Dict = None,
        random_state: Optional[int] = None,
        max_total_time: int = 3600,
        max_eval_time: Optional[int] = None,
        n_jobs: Optional[int] = None,
        max_memory_mb: Optional[int] = None,
        verbosity: int = logging.WARNING,
        search: BaseSearch = AsyncEA(),
        post_processing: BasePostProcessing = BestFitPostProcessing(),
        output_directory: Optional[str] = None,
        store: str = "logs",
    ):
        """

        Parameters
        ----------
        scoring: str, Metric or Tuple
            Specifies the/all metric(s) to optimize towards.
            A string will be converted to Metric.
            A tuple must specify each metric with the same type (e.g. all str).
            See :ref:`Metrics` for built-in metrics.

        regularize_length: bool (default=True)
            If True, add pipeline length as an optimization metric.
            Short pipelines should then be preferred over long ones.

        max_pipeline_length: int, optional (default=None)
            If set, limit the maximum number of steps in any evaluated pipeline.
            Encoding and imputation are excluded.

        config: Dict
            Specifies available components and their valid hyperparameter settings.
            For more information, see :ref:`search_space_configuration`.

        random_state:  int, optional (default=None)
            Seed for the random number generators used in the process.
            However, with `n_jobs > 1`,
            there will be randomization introduced by multi-processing.
            For reproducible results, set this and use `n_jobs=1`.

        max_total_time: positive int (default=3600)
            Time in seconds that can be used for the `fit` call.

        max_eval_time: positive int, optional (default=None)
            Time in seconds that can be used to evaluate any one single individual.
            If None, set to 0.1 * max_total_time.

        n_jobs: int, optional (default=None)
            The amount of parallel processes that may be created to speed up `fit`.
            Accepted values are positive integers, -1 or None.
            If -1 is specified, multiprocessing.cpu_count() processes are created.
            If None is specified, multiprocessing.cpu_count() / 2 processes are created.

        max_memory_mb: int, optional (default=None)
            Sets the total amount of memory GAMA is allowed to use (in megabytes).
            If not set, GAMA will use as much as it needs.
            GAMA is not guaranteed to respect this limit at all times,
            but it should never violate it for too long.

        verbosity: int (default=logging.WARNING)
            Sets the level of log messages to be automatically output to terminal.

        search: BaseSearch (default=AsyncEA())
            Search method to use to find good pipelines. Should be instantiated.

        post_processing: BasePostProcessing (default=BestFitPostProcessing())
            Post-processing method to create a model after the search phase.
            Should be an instantiated subclass of BasePostProcessing.

        output_directory: str, optional (default=None)
            Directory to use to save GAMA output. This includes both intermediate
            results during search and logs.
            If set to None, generate a unique name ("gama_HEXCODE").

        store: str (default='logs')
            Determines which data is stored after each run:
             - 'nothing': keep nothing from this run
             - 'models': keep only cache with models and predictions
             - 'logs': keep only the logs
             - 'all': keep logs and cache with models and predictions
        """
        if not output_directory:
            output_directory = f"gama_{str(uuid.uuid4())}"
        self.output_directory = os.path.abspath(os.path.expanduser(output_directory))
        if not os.path.exists(self.output_directory):
            os.mkdir(self.output_directory)

        register_stream_log(verbosity)
        if store in ["logs", "all"]:
            log_file = os.path.join(self.output_directory, "gama.log")
            log_handler = logging.FileHandler(log_file)
            log_handler.setLevel(logging.DEBUG)
            log_format = logging.Formatter("[%(asctime)s - %(name)s] %(message)s")
            log_handler.setFormatter(log_format)
            logging.getLogger("gama").addHandler(log_handler)

        arguments = ",".join(
            [
                f"{k}={v}"
                for (k, v) in locals().items()
                if k not in ["self", "config", "log_file", "log_handler", "log_format"]
            ]
        )
        log.info(f"Using GAMA version {__version__}.")
        log.info(f"INIT:{self.__class__.__name__}({arguments})")

        if n_jobs is None:
            n_jobs = multiprocessing.cpu_count() // 2
            log.debug("n_jobs defaulted to %d", n_jobs)

        err = ""
        if max_total_time is None or max_total_time <= 0:
            err = f"Expect positive int for max_total_time, got {max_total_time}."
        if max_eval_time is not None and max_eval_time <= 0:
            err = f"Expect None or positive int for max_eval_time, got {max_eval_time}."
        if n_jobs < -1 or n_jobs == 0:
            err = f"n_jobs should be -1 or positive int but is {n_jobs}."
        if err:
            self.cleanup("all")
            raise ValueError(err)

        setattr(
            AsyncEvaluator,
            "__init__",
            partialmethod(
                AsyncEvaluator.__init__,
                n_workers=multiprocessing.cpu_count() if n_jobs is None else n_jobs,
                memory_limit_mb=max_memory_mb,
                logfile=os.path.join(self.output_directory, "memory.log"),
            ),
        )

        if max_eval_time is None:
            max_eval_time = round(0.1 * max_total_time)
        if max_eval_time > max_total_time:
            log.warning(
                f"max_eval_time ({max_eval_time}) > max_total_time ({max_total_time}) "
                f"is not allowed. max_eval_time set to {max_total_time}."
            )
            max_eval_time = max_total_time

        self._max_eval_time = max_eval_time
        self._time_manager = TimeKeeper(max_total_time)
        self._metrics: Tuple[Metric, ...] = scoring_to_metric(scoring)
        self._regularize_length = regularize_length
        self._search_method: BaseSearch = search
        self._post_processing = post_processing
        self._store = store

        if random_state is not None:
            random.seed(random_state)
            np.random.seed(random_state)

        self._x: Optional[pd.DataFrame] = None
        self._y: Optional[pd.DataFrame] = None
        self._basic_encoding_pipeline: Optional[Pipeline] = None
        self._fixed_pipeline_extension: List[Tuple[str, TransformerMixin]] = []
        self._inferred_dtypes: List[Type] = []
        self.model: object = None
        self._final_pop: List[Individual] = []

        self._subscribers: Dict[str, List[Callable]] = defaultdict(list)
        cache_directory = os.path.join(self.output_directory, "cache")
        if isinstance(post_processing, EnsemblePostProcessing):
            self._evaluation_library = EvaluationLibrary(
                m=post_processing.hyperparameters["max_models"],
                n=post_processing.hyperparameters["hillclimb_size"],
                cache=cache_directory,
            )
        else:
            # Don't keep memory-heavy evaluation meta-data (predictions, estimators)
            self._evaluation_library = EvaluationLibrary(m=0, cache=cache_directory)
        self.evaluation_completed(self._evaluation_library.save_evaluation)
        e = search.logger(os.path.join(self.output_directory, "evaluations.log"))
        self.evaluation_completed(e.log_evaluation)

        self._pset, parameter_checks = pset_from_config(config)

        max_start_length = 3 if max_pipeline_length is None else max_pipeline_length
        self._operator_set = OperatorSet(
            mutate=partial(
                random_valid_mutation_in_place,
                primitive_set=self._pset,
                max_length=max_pipeline_length,
            ),
            mate=partial(random_crossover, max_length=max_pipeline_length),
            create_from_population=partial(create_from_population, cxpb=0.2, mutpb=0.8),
            create_new=partial(
                create_random_expression,
                primitive_set=self._pset,
                max_length=max_start_length,
            ),
            compile_=compile_individual,
            eliminate=eliminate_from_pareto,
            evaluate_callback=self._on_evaluation_completed,
            completed_evaluations=self._evaluation_library.lookup,
        )

    def cleanup(self, which="evaluations"):
        cache_directory = os.path.join(self.output_directory, "cache")
        if not os.path.exists(self.output_directory):
            return  # Cleanup has been called previously

        if which in ["logs", "all"]:
            for file in os.listdir(self.output_directory):
                if file.endswith(".log"):
                    os.remove(os.path.join(self.output_directory, file))
        if which in ["evaluations", "all"] and os.path.exists(cache_directory):
            shutil.rmtree(cache_directory)
        if which == "all":
            os.rmdir(self.output_directory)

    def _np_to_matching_dataframe(self, x: np.ndarray) -> pd.DataFrame:
        """ Format np array to dataframe whose column types match the training data. """
        if not isinstance(x, np.ndarray):
            raise TypeError(f"Expected x to be of type 'numpy.ndarray' not {type(x)}.")

        x = pd.DataFrame(x)
        for i, dtype in enumerate(self._inferred_dtypes):
            x[i] = x[i].astype(dtype)
        return x

    def _prepare_for_prediction(self, x):
        if isinstance(x, np.ndarray):
            x = self._np_to_matching_dataframe(x)
        x = self._basic_encoding_pipeline.transform(x)
        return x

    def _predict(self, x: pd.DataFrame):
        raise NotImplementedError("_predict is implemented by base classes.")

    def predict(self, x: Union[pd.DataFrame, np.ndarray]):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            A dataframe or array with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            array with predictions of shape (N,) where N is len(x)
        """
        x = self._prepare_for_prediction(x)
        return self._predict(x)

    def predict_from_file(
        self,
        file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
        **kwargs,
    ) -> np.ndarray:
        """ Predict the target for input found in the ARFF file.

        Parameters
        ----------
        file_path: str
            A csv or ARFF file with the same columns as the one that used in fit.
            Target column must be present in file, but its values are ignored.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the ARFF file.
        **kwargs:
            Any additional arguments for calls to pandas.read_csv or arff.load.

        Returns
        -------
        numpy.ndarray
            array with predictions for each row in the ARFF file.
        """
        x, _ = X_y_from_file(
            file_path, split_column=target_column, encoding=encoding, **kwargs
        )
        x = self._prepare_for_prediction(x)
        return self._predict(x)

    def score(
        self, x: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray]
    ) -> float:
        """ Calculate `self.scoring` metric of the model on (x, y).

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            Data to predict target values for.
        y: pandas.Series or numpy.ndarray
            True values for the target.

        Returns
        -------
        float
            The score obtained on the given test data according to the `scoring` metric.
        """
        predictions = (
            self.predict_proba(x)  # type: ignore
            if self._metrics[0].requires_probabilities
            else self.predict(x)
        )
        return self._metrics[0].score(y, predictions)

    def score_from_file(
        self,
        file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
        **kwargs,
    ) -> float:
        """ Calculate `self.scoring` metric of the model on data in the file.

        Parameters
        ----------
        file_path: str
            A csv or ARFF file with which to calculate the score.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the ARFF file.
        **kwargs:
            Any additional arguments for calls to pandas.read_csv or arff.load.

        Returns
        -------
        float
            The score obtained on the given test data according to the `scoring` metric.
        """
        x, y = X_y_from_file(
            file_path, split_column=target_column, encoding=encoding, **kwargs
        )
        return self.score(x, y)

    def fit_from_file(
        self,
        file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
        warm_start: Optional[List[Individual]] = None,
        **kwargs,
    ) -> None:
        """ Find and fit a model to predict the target column (last) from other columns.

        Parameters
        ----------
        file_path: str
            Path to a csv or ARFF file containing the training data.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the file.
        warm_start: List[Individual], optional (default=None)
            A list of individual to start the search  procedure with.
            If None is given, random start candidates are generated.
        **kwargs:
            Any additional arguments for calls to pandas.read_csv or arff.load.

        """
        x, y = X_y_from_file(file_path, target_column, encoding, **kwargs)
        self.fit(x, y, warm_start)

    def fit(
        self,
        x: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.DataFrame, pd.Series, np.ndarray],
        warm_start: Optional[List[Individual]] = None,
    ) -> "Gama":
        """ Find and fit a model to predict target y from X.

        Various possible machine learning pipelines will be fit to the (X,y) data.
        Using Genetic Programming, the pipelines chosen should lead to gradually
        better models. Pipelines will internally be validated using cross validation.

        After the search termination condition is met, the best found pipeline
        configuration is then used to train a final model on all provided data.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray, shape = [n_samples, n_features]
            Training data. All elements must be able to be converted to float.
        y: pandas.DataFrame, pandas.Series or numpy.ndarray, shape = [n_samples,]
            Target values.
            If a DataFrame is provided, assumes the first column contains target values.
        warm_start: List[Individual], optional (default=None)
            A list of individual to start the search  procedure with.
            If None is given, random start candidates are generated.
        """

        with self._time_manager.start_activity(
            "preprocessing", activity_meta=["default"]
        ):
            x, self._y = format_x_y(x, y)
            self._inferred_dtypes = x.dtypes
            self._x, self._basic_encoding_pipeline = basic_encoding(x)
            self._fixed_pipeline_extension = basic_pipeline_extension(self._x)
            self._operator_set._safe_compile = partial(
                compile_individual, preprocessing_steps=self._fixed_pipeline_extension
            )

            store_pipelines = (
                self._evaluation_library._m is None or self._evaluation_library._m > 0
            )
            if store_pipelines and self._x.shape[0] * self._x.shape[1] > 6_000_000:
                # if m > 0, we are storing models for each evaluation. For this size
                # KNN will create models of about 76Mb in size, which is too big, so
                # we exclude it from search:
                log.info("Excluding KNN from search because the dataset is too big.")
                from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

                self._pset["prediction"] = [
                    p
                    for p in self._pset["prediction"]
                    if p.identifier not in [KNeighborsClassifier, KNeighborsRegressor]
                ]

            if store_pipelines and self._x.shape[1] > 50:
                log.info("Data has too many features to include PolynomialFeatures")
                from sklearn.preprocessing import PolynomialFeatures

                self._pset["data"] = [
                    p
                    for p in self._pset["data"]
                    if p.identifier not in [PolynomialFeatures]
                ]

        fit_time = int(
            (1 - self._post_processing.time_fraction)
            * self._time_manager.total_time_remaining
        )

        with self._time_manager.start_activity(
            "search",
            time_limit=fit_time,
            activity_meta=[self._search_method.__class__.__name__],
        ):
            self._search_phase(warm_start, timeout=fit_time)

        with self._time_manager.start_activity(
            "postprocess",
            time_limit=int(self._time_manager.total_time_remaining),
            activity_meta=[self._post_processing.__class__.__name__],
        ):
            best_individuals = list(
                reversed(
                    sorted(
                        self._final_pop,
                        key=lambda ind: cast(Fitness, ind.fitness).values,
                    )
                )
            )
            self._post_processing.dynamic_defaults(self)
            self.model = self._post_processing.post_process(
                self._x,
                self._y,
                self._time_manager.total_time_remaining,
                best_individuals,
            )
        to_clean = dict(nothing="all", logs="evaluations", models="logs")
        self.cleanup(to_clean[self._store])
        return self

    def _search_phase(
        self, warm_start: Optional[List[Individual]] = None, timeout: float = 1e6
    ):
        """ Invoke the search algorithm, populate `final_pop`. """
        if warm_start:
            if not all([isinstance(i, Individual) for i in warm_start]):
                raise TypeError("`warm_start` must be a list of Individual.")
            pop = warm_start
        else:
            pop = [self._operator_set.individual() for _ in range(50)]

        deadline = time.time() + timeout

        evaluate_pipeline = partial(
            gama.genetic_programming.compilers.scikitlearn.evaluate_pipeline,
            x=self._x,
            y_train=self._y,
            metrics=self._metrics,
        )
        AsyncEvaluator.defaults = dict(evaluate_pipeline=evaluate_pipeline)

        self._operator_set.evaluate = partial(
            gama.genetic_programming.compilers.scikitlearn.evaluate_individual,
            # evaluate_pipeline=evaluate_pipeline,
            timeout=self._max_eval_time,
            deadline=deadline,
            add_length_to_score=self._regularize_length,
        )

        try:
            with stopit.ThreadingTimeout(timeout):
                self._search_method.dynamic_defaults(self._x, self._y, timeout)
                self._search_method.search(self._operator_set, start_candidates=pop)
        except KeyboardInterrupt:
            log.info("Search phase terminated because of Keyboard Interrupt.")

        self._final_pop = self._search_method.output
        n_evaluations = len(self._evaluation_library.evaluations)
        log.info(f"Search phase evaluated {n_evaluations} individuals.")

    def export_script(
        self, file: Optional[str] = "gama_pipeline.py", raise_if_exists: bool = False
    ):
        """ Export a Python script which sets up the best found pipeline.

        Can only be called after `fit`.

        Example
        -------
        After the AutoML search process has completed (i.e. `fit` has been called),
        the model which has been found by GAMA may be exported to a Python file.
        The Python file will define the found pipeline or ensemble.

        .. code-block:: python

            automl = GamaClassifier()
            automl.fit(X, y)
            automl.export_script('my_pipeline_script.py')

        The resulting script will define a variable `pipeline` or `ensemble`,
        depending on the post-processing method that was used after search.

        Parameters
        ----------
        file: str, optional (default='gama_pipeline.py')
            Desired filename of the exported Python script.
            If None, return the code as str instead, it will not be formatted(!).
        raise_if_exists: bool (default=False)
            If True, raise an error if the file already exists.
            If False, overwrite `file` if it already exists.
        """
        if self.model is None:
            raise RuntimeError(STR_NO_OPTIMAL_PIPELINE)
        if raise_if_exists and file is not None and os.path.isfile(file):
            raise FileExistsError(f"File {file} already exists.")

        if self._basic_encoding_pipeline is not None:
            script_text = self._post_processing.to_code(
                self._basic_encoding_pipeline.steps + self._fixed_pipeline_extension
            )
        else:
            script_text = self._post_processing.to_code(self._fixed_pipeline_extension)

        if file:
            with open(file, "w") as fh:
                fh.write(script_text)
            subprocess.call(["black", file])
        else:
            return script_text

    def _safe_outside_call(self, fn):
        """ Calls fn logging and ignoring all exceptions except TimeoutException. """
        try:
            fn()
        except stopit.utils.TimeoutException:
            raise
        except Exception:
            # We actually want to catch any other exception here,
            # because the callback code can be arbitrary (it can be provided by users).
            # This excuses the catch-all Exception.
            # Note KeyboardInterrupts are not exceptions and get elevated to the caller.
            log.warning("Exception during callback.", exc_info=True)

        if self._time_manager.current_activity.exceeded_limit:
            log.info("Time exceeded during callback, but exception was swallowed.")
            raise stopit.utils.TimeoutException

    def _on_evaluation_completed(self, evaluation: Evaluation):
        for callback in self._subscribers["evaluation_completed"]:
            self._safe_outside_call(partial(callback, evaluation))

    def evaluation_completed(self, callback: Callable[[Evaluation], Any]) -> None:
        """ Register a callback function that is called when an evaluation is completed.

        Parameters
        ----------
        callback: Callable[[Evaluation], Any]
            Function to call when a pipeline is evaluated, return values are ignored.
        """
        self._subscribers["evaluation_completed"].append(callback)



import logging

TIME_FORMAT = "%Y-%m-%d %H:%M:%S,%f"

# We also produce log messages below DEBUG level (machine parsable).
MACHINE_LOG_LEVEL = 5
gama_log = logging.getLogger("gama")
gama_log.setLevel(MACHINE_LOG_LEVEL)



#!/usr/bin/env python
import os

from setuptools import setup, find_packages

with open("gama/__version__.py", "r") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

base = [
    "numpy>=1.14.0",
    "scipy>=1.0.0",
    "scikit-learn>=0.20.0",
    "pandas>=1.0,<1.1",
    "stopit>=1.1.1",
    "liac-arff>=2.2.2",
    "category-encoders>=1.2.8",
    "black==19.10b0",
    "psutil",
]

vis = [
    "dash==1.3",
    "dash-daq==0.1.0",
    "dash-bootstrap-components",
    "visdcc",
]

doc = ["sphinx", "sphinx_rtd_theme"]

test = [
    "pre-commit==2.1.1",
    "pytest>=4.4.0",
    "pytest-mock",
    "pytest-xdist<2.0.0",
    "codecov",
    "pytest-cov",
]

# Black, Flake8 and Mypy will be installed through calling pre-commit install
dev = test + doc
all_ = test + doc + vis

with open(os.path.join("README.md")) as fid:
    README = fid.read()

setup(
    name="gama",
    version=version,
    description="A package for automated machine learning based on scikit-learn.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Pieter Gijsbers",
    author_email="p.gijsbers@tue.nl",
    url="https://github.com/PGijsbers/GAMA",
    project_urls={
        "Bug Tracker": "https://github.com/PGijsbers/gama/issues",
        "Documentation": "https://pgijsbers.github.io/gama/",
        "Source Code": "https://github.com/PGijsbers/gama",
    },
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=base,
    extras_require={"vis": vis, "dev": dev, "all": all_,},
    python_requires=">=3.6.0",
    entry_points={
        "console_scripts": [
            "gama=gama.utilities.cli:main",
            "gamadash=gama.dashboard.app:main",
        ]
    },
)



from typing import List

from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html


def button_header(text: str, id_: str, level: int = 4):
    header = f"{'#' * level} {text}"
    return dbc.FormGroup(
        [dbc.Button([dcc.Markdown(header)], id=id_, block=True, color="primary")]
    )


def markdown_header(text: str, level: int = 4, with_horizontal_rule: bool = True):
    header = f"{'#' * level} {text}"
    # matching length '-' not required but nicer.
    hr = f"\n{'-'*(level + 1 + len(text))}"
    return dcc.Markdown(f"{header}{hr if with_horizontal_rule else ''}")


class CollapsableSection:
    """ A Form with a ButtonHeader which when presses collapses/expands the Form. """

    def __init__(
        self, header: str, controls: List[dbc.FormGroup], start_open: bool = True
    ):
        self._header = header
        self._start_open = start_open
        self._header_id = f"{header}-header"
        self._form_id = f"{header}-form"
        self._controls = controls

        self.html = self._build_content()

    def _build_content(self) -> html.Div:
        form_header = button_header(self._header, id_=self._header_id)
        self.form = dbc.Form(self._controls)
        collapsable_form = dbc.Collapse(
            id=self._form_id, children=[self.form], is_open=self._start_open
        )
        return html.Div([form_header, collapsable_form])

    def register_callbacks(self, app):
        app.callback(
            Output(self._form_id, "is_open"),
            [Input(self._header_id, "n_clicks")],
            [State(self._form_id, "is_open")],
        )(_toggle_collapse)


def _toggle_collapse(click, is_open: bool):
    if click:
        return not is_open
    return is_open



import multiprocessing
import os
from typing import Optional, List, Dict, Tuple, Callable

import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table
from dash.dependencies import Input, Output, State

from gama.dashboard.pages.base_page import BasePage
from gama.data_loading import file_to_pandas, load_feature_metadata_from_file


class HomePage(BasePage):
    callbacks: List[Tuple[Tuple, Callable]] = []

    def __init__(self):
        super().__init__(name="Home", alignment=0)
        self.id = "home-page"

    def build_page(self, app, controller):
        self._build_content(app, controller)
        if app is not None:
            self._register_callbacks(app)

    def _build_content(self, app, controller) -> html.Div:
        """ Build all the components of the page. """
        configuration = build_configuration_menu(app, controller)
        configuration.style["width"] = "35%"
        configuration.style["float"] = "left"
        data_navigator = build_data_navigator()
        data_navigator.style["width"] = "65%"
        data_navigator.style["float"] = "right"
        self._content = html.Div(id=self.id, children=[configuration, data_navigator])
        return self._content

    def _register_callbacks(self, app):
        for (io, fn) in HomePage.callbacks:
            app.callback(*io)(fn)
        HomePage.callbacks = []

    def load(self):
        pass

    def unload(self):
        pass


# === Configuration Menu ===


def cpu_slider():
    n_cpus = multiprocessing.cpu_count()
    id_ = "cpu_slider"
    cpu_input = dbc.FormGroup(
        [
            dbc.Label("N Jobs", html_for=id_, width=6),
            dbc.Col(
                dcc.Slider(
                    id=id_,
                    min=1,
                    max=n_cpus,
                    updatemode="drag",
                    value=1,
                    marks={1: "1", n_cpus: str(n_cpus)},
                    persistence_type="session",
                    persistence=True,
                )
            ),
        ],
        row=True,
    )
    HomePage.callbacks.append(
        (
            (
                Output(id_, "marks"),
                [Input(id_, "value")],
                [State(id_, "min"), State(id_, "max")],
            ),
            update_marks,
        )
    )
    return cpu_input


def time_nud(
    label_text: str,
    hour_id: str,
    hour_default: int,
    minute_id: str,
    minute_default: int,
):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=hour_id, width=6),
            dbc.Col(
                dbc.InputGroup(
                    [
                        dbc.Input(
                            id=hour_id,
                            type="number",
                            min=0,
                            max=99,
                            step=1,
                            value=hour_default,
                        ),
                        dbc.InputGroupAddon("H", addon_type="append"),
                    ]
                )
            ),
            dbc.Col(
                dbc.InputGroup(
                    [
                        dbc.Input(
                            id=minute_id,
                            type="number",
                            min=0,
                            max=59,
                            step=1,
                            value=minute_default,
                        ),
                        dbc.InputGroupAddon("M", addon_type="append"),
                    ]
                )
            ),
        ],
        row=True,
    )


def toggle_button(label_text: str, id_: str, start_on: bool = True):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(
                dbc.Checklist(
                    id=id_,
                    options=[{"label": "", "value": "on"}],
                    switch=True,
                    value="on" if start_on else "off",
                )
            ),
        ],
        row=True,
    )


def text_input(label_text: str, default_text: str, id_: str):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(
                dbc.Input(
                    id=id_, type="text", placeholder=default_text, value=default_text
                )
            ),
        ],
        row=True,
    )


def dropdown(
    label_text: str, id_: str, options: Dict[str, str], value: Optional[str] = None
):
    """ options formatted as {LABEL_KEY: LABEL_TEXT, ...} """
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(
                dcc.Dropdown(
                    id=id_,
                    options=[
                        {"label": text, "value": key} for key, text in options.items()
                    ],
                    clearable=False,
                    value=value,
                    persistence_type="session",
                    persistence=True,
                ),
            ),
        ],
        row=True,
    )


def button_header(text: str, id_: str, level: int = 4):
    header = f"{'#' * level} {text}"
    return dbc.FormGroup(
        [dbc.Button([dcc.Markdown(header)], id=id_, block=True, color="primary")]
    )


def markdown_header(text: str, level: int = 4, with_horizontal_rule: bool = True):
    header = f"{'#' * level} {text}"
    hr = f"\n{'-'*(level + 1 + len(text))}"  # matching length '-' not required but nice
    return dcc.Markdown(f"{header}{hr if with_horizontal_rule else ''}")


def toggle_collapse(click, is_open: bool):
    if click:
        return not is_open
    return is_open


def collapsable_section(
    header: str, controls: List[dbc.FormGroup], start_open: bool = True
):
    header_id = f"{header}-header"
    form_id = f"{header}-form"

    form_header = button_header(header, id_=header_id)
    collapsable_form = dbc.Collapse(
        id=form_id, children=[dbc.Form(controls)], is_open=start_open
    )

    HomePage.callbacks.append(
        (
            (
                Output(form_id, "is_open"),
                [Input(header_id, "n_clicks")],
                [State(form_id, "is_open")],
            ),
            toggle_collapse,
        )
    )
    return form_header, collapsable_form


def build_configuration_menu(app, controller) -> html.Div:
    # Optimization
    from gama.utilities.metrics import all_metrics

    metrics = {m: m.replace("_", " ") for m in all_metrics}
    metrics.update({"default": "default"})
    scoring_input = dropdown(
        "Metric", "metric_dropdown", options=metrics, value="default"
    )
    regularize_input = toggle_button(
        "Prefer short pipelines", "regularize_length_switch"
    )
    optimization = collapsable_section(
        "Optimization", [scoring_input, regularize_input]
    )

    # Resources
    cpu_input = cpu_slider()
    max_total_time_input = time_nud(
        "Max Runtime",
        hour_id="max_total_h",
        hour_default=1,
        minute_id="max_total_m",
        minute_default=0,
    )
    max_eval_time_input = time_nud(
        "Max time per pipeline",
        hour_id="max_eval_h",
        hour_default=0,
        minute_id="max_eval_m",
        minute_default=5,
    )
    resources = collapsable_section(
        "Resources", [cpu_input, max_total_time_input, max_eval_time_input]
    )

    # Advanced
    log_path = text_input("Log Directory", "~/GamaLog", "logpath")
    advanced = collapsable_section("Advanced", [log_path], start_open=False)

    # Go!
    go_button = dbc.Button(
        [dcc.Markdown("#### Go!")],
        id="go-button",
        block=True,
        color="success",
        disabled=True,
    )

    def start_gama(n_click, running_tab_style, *args):
        controller.start_gama(*args)
        running_tab_style["display"] = "inline"
        return "danger", dcc.Markdown("#### Stop!"), "Running", running_tab_style

    app.callback(
        [
            Output("go-button", "color"),
            Output("go-button", "children"),
            Output("page-tabs", "value"),
            Output("Running-tab", "style"),
        ],
        [Input("go-button", "n_clicks")],
        [
            State("Running-tab", "style"),
            State("metric_dropdown", "value"),
            State("regularize_length_switch", "value"),
            State("cpu_slider", "value"),
            State("max_total_h", "value"),
            State("max_total_m", "value"),
            State("max_eval_h", "value"),
            State("max_eval_m", "value"),
            State("file-path-input", "value"),
            State("logpath", "value"),
            State("target_dropdown", "value"),
        ],
    )(start_gama)

    return html.Div(
        children=[
            markdown_header("Configure GAMA", level=2),
            *optimization,
            *resources,
            *advanced,
            go_button,
        ],
        style={"box-shadow": "1px 1px 1px black", "padding": "2%"},
    )


def update_marks(selected_value, min_, max_):
    return {min_: str(min_), selected_value: str(selected_value), max_: str(max_)}


def build_data_navigator() -> html.Div:
    upload_file = dbc.Input(
        id="file-path-input",
        placeholder="Path to data file, e.g. ~/data/mydata.arff",
        type="text",
    )

    modes = ["None", "Small", "All"]
    settings = dbc.FormGroup(
        [
            dbc.Label("Target", html_for="target_dropdown", width=2),
            dbc.Col(
                dcc.Dropdown(
                    id="target_dropdown",
                    options=[{"label": "-", "value": "a"}],
                    clearable=False,
                    value="a",
                    # persistence_type="session",
                    # persistence=True,
                ),
                width=4,
            ),
            dbc.Label("Preview Mode", html_for="preview_dropdown", width=2),
            dbc.Col(
                dcc.Dropdown(
                    id="preview_dropdown",
                    options=[{"label": m, "value": m.lower()} for m in modes],
                    clearable=False,
                    value="none",
                    # persistence_type="session",
                    # persistence=True,
                ),
                width=4,
            ),
        ],
        row=True,
    )

    table_container = html.Div(id="table-container", children=["No data loaded."])

    data_settings = html.Div(
        id="data-settings-container",
        children=[settings, table_container],
        style={"margin": "10px"},
    )

    def update_data_table(filename, mode):
        if filename is not None and os.path.isfile(filename):
            if mode in ["all", "small"]:
                df = file_to_pandas(filename)
                if mode == "small":
                    df = df.head(50)

                data_table = dash_table.DataTable(
                    id="table",
                    columns=[{"name": c, "id": c} for c in df.columns],
                    data=df.to_dict("records"),
                    editable=False,
                    style_table={"maxHeight": "500px", "overflowY": "scroll"},
                )
                attributes = list(df.columns)
            else:
                data_table = "Preview not enabled."
                attributes = list(load_feature_metadata_from_file(filename))

            target_options = [{"label": c, "value": c} for c in attributes]
            default_target = attributes[-1]

            return [data_table], target_options, default_target, False
        return ["No data loaded"], [{"label": "-", "value": "a"}], "a", True

    HomePage.callbacks.append(
        (
            (
                [
                    Output("table-container", "children"),
                    Output("target_dropdown", "options"),
                    Output("target_dropdown", "value"),
                    Output("go-button", "disabled"),
                ],
                [
                    Input("file-path-input", "value"),
                    Input("preview_dropdown", "value"),
                ],
            ),
            update_data_table,
        )
    )

    return html.Div(
        children=[
            markdown_header("Data Navigator", level=2),
            upload_file,
            data_settings,
        ],
        style={"box-shadow": "1px 1px 1px black", "padding": "2%"},
    )



import logging
import sys

gama_log = logging.getLogger("gama")


def register_stream_log(verbosity):
    previously_registered_handler = [
        handler for handler in gama_log.handlers if hasattr(handler, "tag")
    ]
    if len(previously_registered_handler) > 0:
        gama_log.debug(
            "Removing StreamHandlers registered by previous GAMA instance(s)."
        )
        gama_log.handlers = [
            handler
            for handler in gama_log.handlers
            if not (
                hasattr(handler, "tag") and isinstance(handler, logging.StreamHandler)
            )
        ]

    stdout_streamhandler = logging.StreamHandler(sys.stdout)
    stdout_streamhandler.tag = "machine_set"
    stdout_streamhandler.setLevel(verbosity)
    gama_log.addHandler(stdout_streamhandler)



from collections import Sequence
from typing import Tuple, List, Optional, Callable, Any


class ParetoFront(Sequence):
    """ A list of tuples in which no one tuple is dominated by another. """

    def __init__(
        self,
        start_list: Optional[List[Any]] = None,
        get_values_fn: Optional[Callable[[Any], Tuple[Any, ...]]] = None,
    ):
        """
        Parameters
        ----------
        start_list: list, optional (default=None).
            List of items of which to calculate the Pareto front.
        get_values_fn: Callable, optional (default=None)
            Function that takes an item and returns a tuple of values,
            such that each should be maximized.
            If left None, it is assumed that items are already such tuples.
        """
        self._get_values_fn = get_values_fn

        self._front: List[Any] = []
        if start_list:
            for item in start_list:
                self.update(item)

        self._iterator_index = 0

    def _get_item_value(self, item):
        if self._get_values_fn is not None:
            return self._get_values_fn(item)
        else:
            return item

    def update(self, new_item: Any):
        """ Update the Pareto front with new_item if it qualifies.

        Parameters
        ----------
        new_item: Any
            Item to be evaluated for submission to the Pareto front.
            Either a Tuple that matches the arity of items already in the Pareto front,
            or an object from which such a Tuple can be extracted
            with the `get_values_fn` provided on `__init__`.

        Returns
        -------
        bool
            True if the Pareto front is updated, False otherwise.
        """
        if not self._front:
            self._front.append(new_item)
            return True

        values = self._get_item_value(new_item)
        old_arity = len(self._get_item_value(self._front[0]))
        if old_arity != len(values):
            raise ValueError(
                "Arity of added tuple must match that of the ones in the Pareto front. "
                f"Front tuples have arity {len(self._front[0])} and "
                f"new tuple has arity {len(values)}."
            )

        to_remove = []

        for old_item in self._front:
            old_values = self._get_item_value(old_item)
            if all(v1 <= v2 for v1, v2 in zip(values, old_values)):
                # old_item dominates this new_item, no update
                return False
            elif all(v1 >= v2 for v1, v2 in zip(values, old_values)):
                # new_item dominates this old_item
                to_remove.append(old_item)
            # else: Neither dominates nor gets dominated by old_item

        # new_item was not dominated by any Pareto-front item, update front:
        for item in to_remove:
            self._front.remove(item)
        self._front.append(new_item)
        return True

    def clear(self):
        """ Removes all items from the Pareto front."""
        self._front = []

    def __len__(self):
        return len(self._front)

    def __getitem__(self, item):
        return self._front[item]

    def __str__(self):
        return str(self._front)

    def __repr__(self):
        if self._get_values_fn is not None:
            fn_name = f", get_values_fn = '{self._get_values_fn.__name__}"
        else:
            fn_name = ""
        return f"ParetoFront({self._front}{fn_name})"



import base64
import itertools
import os
import shutil
import uuid
from typing import Dict, List, Optional

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State

from gama.dashboard.pages.base_page import BasePage
from gama.logging.GamaReport import GamaReport
from ..plotting import plot_preset_graph


class AnalysisPage(BasePage):
    def __init__(self):
        super().__init__(name="Analysis", alignment=-1)
        self.id = "analysis-page"
        self.reports: Dict[str, GamaReport] = {}

    def build_page(self, app, controller):
        upload_box = html.Div(
            id="upload-container",
            children=[
                dcc.Upload(
                    id="upload-box",
                    children=html.Div([html.A("Select or drop log(s).")]),
                    style={
                        "width": "100%",
                        "height": "60px",
                        "lineHeight": "60px",
                        "borderWidth": "1px",
                        "borderStyle": "dashed",
                        "borderRadius": "5px",
                        "textAlign": "center",
                        "display": "inline-block",
                    },
                    multiple=True,
                )
            ],
            style=dict(
                width=f'{len("Select or drop log(s).")}em',
                display="inline-block",
                float="right",
            ),
        )

        # Top

        presets = [
            {"label": "#Pipeline by learner", "value": "number_pipeline_by_learner"},
            {"label": "#Pipeline by size", "value": "number_pipeline_by_size"},
            {"label": "Best score over time", "value": "best_over_time"},
            {"label": "Best score over iterations", "value": "best_over_n"},
            {"label": "Size vs Metric", "value": "size_vs_metric"},
            {"label": "Evaluation Times", "value": "evaluation_times_dist"},
            {"label": "Evaluations by Rung", "value": "n_by_rung"},
            {"label": "Time by Rung", "value": "time_by_rung"},
            # {"label": "Custom", "value": "custom"},
        ]

        preset_container = html.Div(
            id="preset-container",
            children=[
                html.Div("Visualization Presets"),
                dcc.Dropdown(
                    id="preset-dropdown",
                    options=presets,
                    value="best_over_n",
                    style=dict(width="90%"),
                ),
            ],
            style=dict(width="100%", display="inline-block", float="left"),
        )

        # sep_agg_radio = dcc.RadioItems(
        #     id="sep-agg-radio",
        #     options=[
        #         {"label": "separate", "value": "separate-line"},
        #         {"label": "aggregate", "value": "aggregate"},
        #     ],
        #     value="separate-line",
        #     style={"width": "90%", "display": "inline-block"},
        # )

        # sep_agg_container = html.Div(
        #     id="sep_agg_container",
        #     children=[html.Div("Style"), sep_agg_radio],
        #     style=dict(display="inline-block", width="50%", float="left"),
        # )

        # left

        dashboard_graph = dcc.Graph(id="dashboard-graph")

        # third_width = {"width": "30%", "display": "inline-block"}
        # plot_control_container = html.Div(
        #     id="plot-controls",
        #     children=[
        #         html.Div(
        #             [html.Label("x-axis"), dcc.Dropdown(id="x-axis-metric")],
        #             style=third_width
        #         ),
        #         html.Div(
        #             [html.Label("y-axis"), dcc.Dropdown(id="y-axis-metric")],
        #             style=third_width
        #         ),
        #         html.Div(
        #             [
        #                 html.Label("plot type"),
        #                 dcc.Dropdown(
        #                     id="plot-type",
        #                     options=[
        #                         {"label": "scatter", "value": "markers"},
        #                         {"label": "line", "value": "lines"},
        #                     ],
        #                     value="lines",
        #                 ),
        #             ],
        #             style=third_width,
        #         ),
        #     ],
        #     style=dict(width="80%", display="none"),
        #     hidden=True,
        # )
        #
        # graph_settings_container = html.Div(
        #     id="graph-settings-container", children=[plot_control_container]
        # )

        # graph_update_timer = dcc.Interval(id="update-timer", interval=2 * 1000)  # ms
        #
        # graph_update_trigger = dcc.Store(id="update-trigger")

        visualization_container = html.Div(
            id="visualization-container",
            children=[
                dashboard_graph,
                # graph_settings_container,
                # graph_update_timer,
                # graph_update_trigger,
            ],
            style={"float": "left", "width": "85%"},
        )

        # right

        file_select = dcc.Checklist(id="select-log-checklist")

        report_select_container = html.Div(
            id="report-select-container",
            children=[preset_container, upload_box, file_select],
            style={"width": "14%", "float": "right", "padding-right": "1%"},
        )

        self._content = html.Div(
            id=self.id, children=[visualization_container, report_select_container],
        )

        app.callback(
            Output("select-log-checklist", "options"),
            [Input("upload-box", "contents")],
            [State("upload-box", "filename")],
        )(self.load_logs)

        app.callback(
            Output("dashboard-graph", "figure"),
            [
                Input("select-log-checklist", "value"),
                Input("preset-dropdown", "value"),
            ],
        )(self.update_graph)

        return self._content

    def load_logs(self, list_of_contents, list_of_names):
        # global aggregate_dataframe
        if list_of_contents is not None:
            tmp_dir = f"tmp_{str(uuid.uuid4())}"
            os.makedirs(tmp_dir)
            for content, filename in zip(list_of_contents, list_of_names):
                content_type, content_string = content.split(",")
                decoded = base64.b64decode(content_string).decode("utf-8")
                with open(os.path.join(tmp_dir, filename), "w") as fh:
                    fh.write(decoded)

            report = GamaReport(tmp_dir)
            report_name = report.search_method
            for i in itertools.count():
                if f"{report_name}_{i}" not in self.reports:
                    break
            self.reports[f"{report_name}_{i}"] = report
            shutil.rmtree(tmp_dir)
            return [{"label": logname, "value": logname} for logname in self.reports]
        return []

    def update_graph(self, logs: List[str], preset_value: Optional[str] = None):
        print(logs, preset_value)  # , aggregate, xaxis, yaxis, mode,
        # if preset_value == "custom":
        #     if logs is None or logs == [] or xaxis is None or yaxis is None:
        #         title = "Load and select a log on the right"
        #         plots = []
        #     else:
        #         title = f"{aggregate} plot of {len(logs)} logs"
        #         if aggregate == "separate-line":
        #             plots = [
        #                 individual_plot(reports[log], xaxis, yaxis, mode)
        #                 for log in logs
        #             ]
        #         if aggregate == "aggregate":
        #             plots = aggregate_plot(
        #               [reports[log] for log in logs], xaxis, yaxis
        #             )
        #     return {
        #         "data": plots,
        #         "layout": {
        #             "title": title,
        #             "xaxis": {"title": f"{xaxis}"},
        #             "yaxis": {"title": f"{yaxis}"},
        #             "hovermode": "closest" if mode == "markers" else "x",
        #         },
        #     }
        if logs is not None:
            # filtered_aggregate = aggregate_dataframe[
            #     aggregate_dataframe.filename.isin(logs)
            # ]
            return plot_preset_graph([self.reports[log] for log in logs], preset_value)
        else:
            return {}



from typing import Callable, NamedTuple, Tuple


class Primitive(NamedTuple):
    """ Defines an operator which takes input and produces output.

    E.g. a preprocessing or classification algorithm.
    """

    input: Tuple[str]
    output: str
    identifier: Callable

    def __str__(self):
        """ str: e.g. "FastICA" """
        return self.identifier.__name__

    def __repr__(self):
        return str(self)



import shlex
import subprocess
import threading
import queue
from typing import List

from dash import Dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as html
import visdcc


def enqueue_output(out, queue_: queue.Queue):
    for line in iter(out.readline, b""):
        queue_.put(line)
    out.close()


class CLIWindow:
    """ A Component for Dash App which simulates a console window """

    def __init__(
        self,
        id_: str,
        app: Dash,
        update_interval_s: float = 1.0,
        auto_scroll: bool = True,
    ):
        self._update_interval_s = update_interval_s
        self.auto_scroll = auto_scroll

        self.console_id = f"{id_}-text"
        self.timer_id = f"{id_}-interval"
        self.js_id = f"{id_}-js"
        self.id_ = id_

        self.autoscroll_script = (
            f"document.getElementById('{self.console_id}').scrollTop"
            f" = document.getElementById('{self.console_id}').scrollHeight"
        )
        self.process = None
        self._thread = None
        self._queue = None
        self._lines: List[str] = []

        self.html = self._build_component()
        self._register_callbacks(app)

    def _build_component(self) -> html.Div:
        timer = dcc.Interval(
            id=self.timer_id, interval=self._update_interval_s * 1000, n_intervals=0
        )
        scroller = visdcc.Run_js(id=self.js_id, run="")
        self.console = dcc.Textarea(
            id=self.console_id,
            contentEditable="false",
            style={
                "height": "100%",
                "width": "100%",
                "borderWidth": "1px",
                "borderRadius": "5px",
                "borderStyle": "dashed",
            },
            persistence_type="session",
            persistence=True,
        )
        return html.Div(
            id=self.id_,
            children=[timer, self.console, scroller],
            style={"height": "100%"},
        )

    def _register_callbacks(self, app):
        app.callback(
            [Output(self.console_id, "value"), Output(self.js_id, "run")],
            [Input(self.timer_id, "n_intervals")],
            [State(self.console_id, "value")],
        )(self.update_console)

    def monitor(self, process):
        self.process = process
        # Because there are only blocking reads to the pipe,
        # we need to read them on a separate thread.
        self._queue = queue.Queue()
        self._thread = threading.Thread(
            target=enqueue_output, args=(self.process.stdout, self._queue), daemon=True
        )
        self._thread.start()

    def call(self, command: str):
        cmd = shlex.split(command)
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        self.monitor(process)

    def update_console(self, _, current_text):
        if self.process is None:
            return [None, None]

        # We want to update the text field if there is new output from the process,
        # or if we detect the text value has been reset (due to e.g. switching tabs).
        try:
            line = self._queue.get_nowait()
            self._lines.append(line.decode("utf-8"))
        except queue.Empty:
            # No new message, update only required if value field had been reset.
            if current_text is not None:
                raise PreventUpdate

        script = self.autoscroll_script if self.auto_scroll else ""
        return ["".join(self._lines), script]
